from jsonschema import Draft4Validator, FormatChecker, exceptions
from collections import deque, OrderedDict
import isodate, datetime, json, copy, io
from lxml import etree

##
ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}

xs = "http://www.w3.org/2001/XMLSchema"
lx = "lixi.org.au/schema/appinfo_elements"
li = "lixi.org.au/schema/appinfo_instructions"


class LIXIValidationError(Exception):
    def __init__(self, message, message_instance=None):
        super().__init__(message)
        self.message_instance = message_instance


def validate_json(json_message, json_schema):
    """Validates a LIXI JSON message against a LIXI JSON schema. Not used currently.
    
    Args:
        json_message (:obj:`dict`, required): A python message dict.  
        json_schema (:obj:`dict`, required): A python schema dict.
    
    Result:
        is_valid: True or False.
        validation_message: A short description of the validation message.
    
    Raises:
        ValidationError: Validation errors for the xml file.
    """

    counter = 1
    out_string = ""

    IDs = []
    IDREFs = []
    custom_format_checker = FormatChecker()
    idref_is_true = True

    @custom_format_checker.checks("lixi-year", exceptions.ValidationError)
    def format_year(string_value):
        try:
            dt = datetime.datetime.strptime(string_value, "%Y")
            return True
        except ValueError as e:
            raise exceptions.ValidationError(
                "'" + string_value + "' is not a valid year"
            )

    @custom_format_checker.checks("lixi-fulldate", exceptions.ValidationError)
    def format_fullyear(string_value):
        try:
            dt = datetime.datetime.strptime(string_value, "%Y-%m-%d")
            return True
        except ValueError as e:
            raise exceptions.ValidationError(
                "'" + string_value + "' is not a valid date"
            )

    @custom_format_checker.checks("lixi-datetime", exceptions.ValidationError)
    def format_datetime(string_value):
        try:
            dt = isodate.parse_datetime(string_value)
            return True
        except ValueError as e:
            raise exceptions.ValidationError(
                "'" + string_value + "' is not a valid ISO date string"
            )

    @custom_format_checker.checks("lixi-ID", exceptions.ValidationError)
    def format_ID(string_value):

        if string_value in IDs:
            raise exceptions.ValidationError("ID '" + string_value + "' is not unique")

        IDs.append(string_value)
        return True

    @custom_format_checker.checks("lixi-IDREF")
    def format_ID(string_value):
        IDREFs.append(string_value)
        return True

    def print_nested_error(error, old_path_list):
        path_list = deque()
        if isinstance(error, exceptions.ValidationError):
            if len(error.context) == 0:
                if isinstance(path_list, deque) and isinstance(error.path, deque):
                    path_list.extend(old_path_list)
                    path_list.extend(error.path)

                result_path = (
                    repr(path_list)
                    .replace("deque(['", "")
                    .replace("'])", "")
                    .replace("', '", ".")
                )
                out_string += (
                    "\nERROR "
                    + repr(counter)
                    + ": "
                    + error.message
                    + "\n  PATH: "
                    + result_path
                )
                counter += 1
            else:
                if isinstance(path_list, deque) and isinstance(error.path, deque):
                    path_list.extend(old_path_list)
                    path_list.extend(error.path)

                for child_error in error.context:
                    print_nested_error(child_error, path_list)

    validator = Draft4Validator(json_schema, format_checker=custom_format_checker)

    if validator.is_valid(json_message):

        for idref in IDREFs:
            if idref not in IDs:
                idref_is_true = False

        if idref_is_true:
            return True, "Message is Valid"
        else:
            out_string += "\n-> Message is Invalid"
            for idref in IDREFs:
                check = False
                if idref not in IDs:
                    check = True
                if check:
                    out_string += (
                        "\nERROR "
                        + repr(counter)
                        + ": IDREF '"
                        + idref
                        + "' does not have a referenced ID, PATH: Message-wide"
                    )
                counter += 1

            return False, out_string

    else:
        out_string += "\n-> Message is Invalid"

        message_errors = []
        message_errors = validator.iter_errors(json_message)

        for error in message_errors:
            path_list = deque()
            print_nested_error(error, path_list)

        for idref in IDREFs:
            check = False
            if idref not in IDs:
                check = True
            if check:
                out_string += (
                    "\nERROR "
                    + repr(counter)
                    + ": IDREF '"
                    + idref
                    + "' does not have a referenced ID, PATH: Message-wide"
                )
            counter += 1

        return False, out_string


def convert_to_json_schema_string(xmlstring):
    """Converts a LIXI XML schema to JSON schema.
    
    Args: 
        xml_schema (:obj:`str`, required):  A lxml etree LIXI schema.
    
    Result:
        A json LIXI schema string.

    """
    xsd_schema = etree.fromstring(xmlstring)
    json_schema = convert_to_json_schema(xsd_schema)

    return json.dumps(json_schema, sort_keys=True, indent=4, ensure_ascii=False)


def convert_to_json_schema(xml_schema):
    """Converts a LIXI XML schema to JSON schema.
    
    Args: 
        xml_schema (:obj:`lxml.etree`, required):  A lxml etree LIXI schema.
    
    Result:
        A json LIXI schema Python dict.

    """
    date_regex = "^(?:[1-9]\\d{3}-(?:(?:0[1-9]|1[0-2])-(?:0[1-9]|1\\d|2[0-8])|(?:0[13-9]|1[0-2])-(?:29|30)|(?:0[13578]|1[02])-31)|(?:[1-9]\\d(?:0[48]|[2468][048]|[13579][26])|(?:[2468][048]|[13579][26])00)-02-29)?$"
    xml_to_json_schema = {
        "string": {"type": "string"},
        "token": {"type": "string"},
        "ID": {
            "type": "string",
            "pattern": "^[a-zA-Z_][a-zA-Z0-9-_.]*$",
            "format": "lixi-ID",
        },
        "IDREF": {
            "type": "string",
            "pattern": "^[a-zA-Z_][a-zA-Z0-9-_.]*$",
            "format": "lixi-IDREF",
        },
        "base64Binary": {
            "pattern": "^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$",
            "type": "string",
        },
        "decimal": {"type": "number"},
        "integer": {"type": "integer"},
        "double": {"type": "number"},
        "dateTime": {"type": "string", "format": "lixi-datetime"},
        "date": {
            "type": "string",
            "pattern": "^(?:[1-9]\\d{3}-(?:(?:0[1-9]|1[0-2])-(?:0[1-9]|1\\d|2[0-8])|(?:0[13-9]|1[0-2])-(?:29|30)|(?:0[13578]|1[02])-31)|(?:[1-9]\\d(?:0[48]|[2468][048]|[13579][26])|(?:[2468][048]|[13579][26])00)-02-29)?$",
            "format": "lixi-date",
        },
        "gYear": {"type": "string", "format": "lixi-year"},
    }

    type_mappings = xml_to_json_schema

    def xsd_to_json_type(element_type, schema, definitions):
        """
        Where base type is found, map it to reserved types provided by default in XSD and if key is 
        being added for the first time add it to definitions as well. for now its thats not possible, 
        in worst case hardcode to string value
        
        @param element_type: a string with 
        @param schema: reference to the parent schema dict
        @param definitions: reference to the difinitions dict
        """

        # Removing the substring xs: from elements
        if "xs:" in element_type:
            element_type = element_type[3:]

        if element_type in definitions:
            schema["$ref"] = "#/definitions/" + element_type
        else:
            if element_type in type_mappings:
                definitions[element_type] = copy.deepcopy(type_mappings[element_type])

                schema["$ref"] = "#/definitions/" + element_type
            else:
                # print('The curious case of an element being referenced before put in definitions')
                schema["$ref"] = "#/definitions/" + element_type

    def add_annotation(element, schema):
        """
        Rules for parsing the xs:annotation element
    
        @param element: an xml parent element with at least one child xs:annotation
        @param schema: reference to the parent schema dict
        """

        if element.find("./xs:annotation/xs:documentation", namespaces=ns) is not None:
            schema["description"] = element.find(
                "./xs:annotation/xs:documentation", namespaces=ns
            ).text

        if (
            element.find("./xs:annotation/xs:appinfo/lx:path", namespaces=ns)
            is not None
        ):
            schema["path"] = element.find(
                "./xs:annotation/xs:appinfo/lx:path", namespaces=ns
            ).text

        if (
            element.find("./xs:annotation/xs:appinfo/lx:label", namespaces=ns)
            is not None
        ):
            schema["label"] = element.find(
                "./xs:annotation/xs:appinfo/lx:label", namespaces=ns
            ).text

        if (
            element.find("./xs:annotation/xs:appinfo/lx:references", namespaces=ns)
            is not None
        ):
            temp_array = []
            for target in element.findall(
                "./xs:annotation/xs:appinfo/lx:references/lx:target", namespaces=ns
            ):
                temp_array.append(target.text)

            schema["references"] = temp_array

        if (
            element.find(
                "./xs:annotation/xs:appinfo/lx:CustomDocumentation", namespaces=ns
            )
            is not None
        ):
            schema["custom_description"] = element.find(
                "./xs:annotation/xs:appinfo/lx:CustomDocumentation", namespaces=ns
            ).text

        if (
            element.find("./xs:annotation/xs:appinfo/lx:CustomExcerpt", namespaces=ns)
            is not None
        ):
            schema["custom_excerpt"] = element.find(
                "./xs:annotation/xs:appinfo/lx:CustomExcerpt", namespaces=ns
            ).text

    def add_attribute(element, parent_schema, passing_schema, definitions):
        """
        Rules for parsing the xs:attribute element

        @param element: an xml parent element with at least one child xs:attribute
        @param schema: reference to the parent schema dict
        @param definitions: reference to the difinitions dict
        """
        for attribute in element.findall("./xs:attribute", ns):
            temp_schema = OrderedDict()

            add_annotation(attribute, temp_schema)

            element_type = attribute.attrib["type"]
            xsd_to_json_type(element_type, temp_schema, definitions)

            if "use" in attribute.attrib:
                if attribute.attrib["use"] != "optional":
                    if "required" not in parent_schema:
                        parent_schema["required"] = []
                    parent_schema["required"].append("@" + attribute.attrib["name"])

            if "fixed" in attribute.attrib:
                temp_schema.pop("$ref", None)
                temp_schema["type"] = "string"
                temp_schema["enum"] = []
                temp_schema["enum"].append(attribute.attrib["fixed"])

            passing_schema["@" + attribute.attrib["name"]] = temp_schema

    def add_complexType(complex_type, parent_schema, passing_schema, definitions):
        """
        Rules for parsing the xs:complexType element
        
        @param complex_type: an xml element with tag xs:complexType
        @param parent_schema: reference to the parent schema of the passing schema
        @param passing_schema: reference to the parent schema of the passing schema
        @param definitions: reference to the difinitions dict
        """
        if complex_type.findall("./xs:attribute", namespaces=ns) is not None:
            add_attribute(complex_type, parent_schema, passing_schema, definitions)

    def add_simpleType(simple_type, schema, definitions):
        """
        Rules for parsing the xs:simpleType element
        
        @param simple_type: an xml element with tag xs:simpleType
        @param schema: reference to the parent schema dict
        @param definitions: reference to the difinitions dict
        """
        temp_schema = OrderedDict()

        add_annotation(simple_type, temp_schema)

        if simple_type.findall("./xs:restriction", namespaces=ns) is not None:
            add_restriction(simple_type, temp_schema, definitions)

        definitions[simple_type.attrib["name"]] = temp_schema

    def add_restriction(element, schema, definitions):
        """
        Rules for parsing restriction from an element
    
        @param element: an xml parent element with at least one child xs:restriction
        @param schema: reference to the parent schema dict
        @param definitions: reference to the difinitions dict
        """
        element_type = element.find("xs:restriction", ns).attrib["base"]
        xsd_to_json_type(element_type, schema, definitions)

        if element.find("./xs:restriction/xs:enumeration", namespaces=ns) is not None:
            add_enumeration(element, schema)

        if element.find("./xs:restriction/xs:pattern", namespaces=ns) is not None:
            add_pattern(element, schema)

        if (
            len(
                element.xpath(
                    "./xs:restriction/xs:minInclusive|./xs:restriction/xs:maxInclusive|./xs:restriction/xs:minExclusive|./xs:restriction/xs:maxExclusive",
                    namespaces=ns,
                )
            )
            > 0
        ):
            add_range(element, schema)

    def add_enumeration(element, schema):
        """
        Rules for parsing enum in an element
        
        @param element: an xml parent element with at least one child xs:enumeration
        @param schema: reference to the parent schema dict
        """
        array = []
        array_defs = {}

        for enum in element.findall("./xs:restriction/xs:enumeration", ns):

            val = enum.attrib["value"]
            array.append(val)

            temp_dict = {}

            if enum.find("./xs:annotation/xs:documentation", namespaces=ns) is not None:
                temp_dict["description"] = enum.find(
                    "./xs:annotation/xs:documentation", namespaces=ns
                ).text

            if (
                enum.find("./xs:annotation/xs:appinfo/lx:path", namespaces=ns)
                is not None
            ):
                temp_dict["path"] = enum.find(
                    "./xs:annotation/xs:appinfo/lx:path", namespaces=ns
                ).text

            if enum.find("./xs:annotation/xs:documentation", namespaces=ns) is not None:
                temp_dict["label"] = enum.find(
                    "./xs:annotation/xs:appinfo/lx:label", namespaces=ns
                ).text

            if len(temp_dict) > 0:
                array_defs[val] = temp_dict

        schema.pop("$ref", None)
        if len(array_defs) > 0:
            schema["enum_definitions"] = array_defs
        schema["enum"] = array

    def add_pattern(element, schema):
        """
        Rules for parsing Pattern in an element
        
        @param element: an xml parent element with at least one child xs:enumeration
        @param schema: reference to the parent schema dict
        """
        # TODO: XML RegEx is just being copied, IT IS NOT SIMILAR to js regex. needs conversion if poss
        schema.pop("$ref", None)
        schema["type"] = "string"
        schema["pattern"] = (
            "^" + element.find("./xs:restriction/xs:pattern", ns).attrib["value"] + "$"
        )

    def add_range(element, schema):
        """
        Rules for parsing integer range in an element
        
        @param element: an xml parent element with at least one child ./xs:restriction/xs:minInclusive|./xs:restriction/xs:maxInclusive|./xs:restriction/xs:minExclusive|./xs:restriction/xs:maxExclusive
        @param schema: reference to the parent schema dict
        """
        # TODO: XML RegEx is just being copied, IT IS NOT SIMILAR to js regex. needs conversion if poss
        schema.pop("$ref", None)

        element_type = element.find("xs:restriction", ns).attrib["base"]
        if element_type == "xs:integer":
            schema["type"] = "integer"
        else:
            schema["type"] = "number"

        if element.find("./xs:restriction/xs:minInclusive", namespaces=ns) is not None:
            schema["minimum"] = element.find(
                "./xs:restriction/xs:minInclusive", namespaces=ns
            ).attrib["value"]

        if element.find("./xs:restriction/xs:maxInclusive", namespaces=ns) is not None:
            schema["maximum"] = element.find(
                "./xs:restriction/xs:maxInclusive", namespaces=ns
            ).attrib["value"]

        if element.find("./xs:restriction/xs:minExclusive", namespaces=ns) is not None:
            schema["exclusiveMinimum"] = element.find(
                "./xs:restriction/xs:minExclusive", namespaces=ns
            ).attrib["value"]

        if element.find("./xs:restriction/xs:maxExclusive", namespaces=ns) is not None:
            schema["exclusiveMaximum"] = element.find(
                "./xs:restriction/xs:maxExclusive", namespaces=ns
            ).attrib["value"]

    def add_element(element, parent_schema, properties_schema, isComplex, definitions):
        """
        recursively parse sub elements
        
        @param element: xml.etree.ElementTree type element to step into
        @param parent_schema: dict to store sub schema values
        @param properties_schema: dict to store sub schema values
        """
        temp_schema = OrderedDict()

        element_name = element.attrib.get("name")
        min_occurs = element.attrib.get("minOccurs", "1")
        max_occurs = element.attrib.get("maxOccurs", "1")

        if max_occurs == "1":
            add_annotation(element, temp_schema)
            temp_schema_properties = OrderedDict()

            # add the required tag in the parent_schema else dont add it
            if min_occurs != "0" and not isComplex:
                if "required" not in parent_schema:
                    parent_schema["required"] = []
                parent_schema["required"].append(element_name)

            if element.attrib.get("type", "object") == "object":
                temp_schema["type"] = "object"

                if (
                    element.find("./xs:complexType", namespaces=ns) is not None
                    or isComplex is True
                ):

                    if isComplex:
                        complex_type = element
                    else:
                        complex_type = element.find("./xs:complexType", namespaces=ns)

                    add_complexType(
                        complex_type, temp_schema, temp_schema_properties, definitions
                    )

                    if (
                        complex_type.find("./xs:choice/xs:element", namespaces=ns)
                        is not None
                    ):
                        
                        if temp_schema_properties:
                            # temp_schema["properties"] = {}
                            # temp_schema["properties"] = copy.deepcopy(temp_schema_properties)                            
                            # temp_schema["properties"]["oneOf"] = []
                            temp_schema["oneOf"] = []

                        else:
                            temp_schema["oneOf"] = []

                        # create a dummy object for every elemenet so that its name can come in
                        for sub_element in complex_type.findall(
                            "./xs:choice/xs:element", namespaces=ns
                        ):
                            new_temp_dict = OrderedDict()
                            new_temp_dict["type"] = "object"
                            new_temp_dict["additionalProperties"] = False
                            new_temp_dict["properties"] = OrderedDict()
                            new_temp_dict["properties"].update(copy.deepcopy(temp_schema_properties))  

                            # handles a case where attributes come along with choice, never for elements - in this case the attr are added to ALL subsequent elements as oneOf cant have attr at same level
                            #if temp_schema_properties:
                                #new_temp_dict["properties"] = copy.deepcopy(
                                    #temp_schema_properties
                                #)

                            # Recurse
                            new_temp_dict["properties"][
                                sub_element.attrib.get("name")
                            ] = add_element(
                                sub_element,
                                new_temp_dict,
                                new_temp_dict["properties"],
                                False,
                                definitions,
                            )
                            
                            if temp_schema_properties:
                                # temp_schema["properties"]["oneOf"].append(new_temp_dict)
                                temp_schema["oneOf"].append(new_temp_dict)
                            else:
                                temp_schema["oneOf"].append(new_temp_dict)                            
                            
                            

                    else:
                        # cater to a single case where xml namespaces arent taken as invalid message
                        # if element_name == "Package":
                        #     temp_schema["additionalProperties"] = True
                        # else:
                        #     temp_schema["additionalProperties"] = False

                        temp_schema["additionalProperties"] = False

                        temp_schema["properties"] = temp_schema_properties

                        if (
                            complex_type.find("./xs:sequence/xs:element", namespaces=ns)
                            is not None
                        ):
                            for sub_element in complex_type.findall(
                                "./xs:sequence/xs:element", namespaces=ns
                            ):
                                # Recurse
                                temp_schema_properties[
                                    sub_element.attrib.get("name")
                                ] = add_element(
                                    sub_element,
                                    temp_schema,
                                    temp_schema_properties,
                                    False,
                                    definitions,
                                )
            else:
                # Recursion end conditions
                if element_name == "InlineAttachment" or element_name == "Comment":
                    temp_schema["type"] = "object"
                    temp_schema["additionalProperties"] = False
                    temp_schema["properties"] = OrderedDict()
                    temp_schema["properties"]["$"] = OrderedDict()
                    xsd_to_json_type(
                        element.attrib.get("type", "object"),
                        temp_schema["properties"]["$"],
                        definitions,
                    )

                else:
                    temp_schema.pop("id", None)
                    xsd_to_json_type(
                        element.attrib.get("type", "object"), temp_schema, definitions
                    )

        else:

            # as an array
            temp_array_schema = OrderedDict()

            add_annotation(element, temp_array_schema)
            temp_array_schema["type"] = "array"
            temp_array_schema["additionalItems"] = False

            temp_array_schema["items"] = OrderedDict()
            temp_array_schema["minItems"] = int(min_occurs)

            if min_occurs == "1" and not isComplex:
                if "required" not in parent_schema:
                    parent_schema["required"] = []
                parent_schema["required"].append(element_name)

            if "unbounded" not in max_occurs:
                temp_array_schema["maxItems"] = int(max_occurs)

            if element.attrib.get("type", "object") == "object":
                new_temp_dict = OrderedDict()
                new_temp_dict["type"] = "object"
                new_temp_dict["additionalProperties"] = False
                new_temp_dict["properties"] = OrderedDict()

                if element.find("./xs:complexType", namespaces=ns) is not None:
                    complex_type = element.find("./xs:complexType", namespaces=ns)
                    add_complexType(
                        complex_type,
                        new_temp_dict,
                        new_temp_dict["properties"],
                        definitions,
                    )

                    if (
                        complex_type.find("./xs:sequence/xs:element", namespaces=ns)
                        is not None
                    ):
                        for sub_element in complex_type.findall(
                            "./xs:sequence/xs:element", namespaces=ns
                        ):
                            # Recurse
                            new_temp_dict["properties"][
                                sub_element.attrib.get("name")
                            ] = add_element(
                                sub_element,
                                new_temp_dict,
                                new_temp_dict["properties"],
                                False,
                                definitions,
                            )

                temp_array_schema["items"] = new_temp_dict

            else:
                # Recursion end conditions
                if element_name == "InlineAttachment" or element_name == "Comment":
                    new_temp_dict = OrderedDict()
                    new_temp_dict["type"] = "object"
                    new_temp_dict["additionalProperties"] = False
                    new_temp_dict["properties"] = OrderedDict()
                    new_temp_dict["properties"]["$"] = OrderedDict()
                    xsd_to_json_type(
                        element.attrib.get("type", "object"),
                        new_temp_dict["properties"]["$"],
                        definitions,
                    )

                    temp_array_schema["items"] = new_temp_dict

                else:
                    new_temp_dict = OrderedDict()
                    xsd_to_json_type(
                        element.attrib.get("type", "object"), new_temp_dict, definitions
                    )
                    temp_array_schema["items"] = new_temp_dict

            # temp_schema['oneOf'].append(temp_array_schema)
            temp_schema = temp_array_schema
        return temp_schema

    # objects that will contain the converted schema
    schema = OrderedDict()
    properties = OrderedDict()
    definitions = (
        OrderedDict()
    )  # we cant make these ordered as the validator would not find compile some references before they are used.
    root = xml_schema

    # This is specified for LIXI XSD, as "Package" is an object, a more generalised approach would only contain the schema tag and definitions tag
    schema["$schema"] = "http://json-schema.org/schema#"
    schema["type"] = "object"
    schema["properties"] = properties
    schema["definitions"] = definitions
    schema["schemadetail"] = {}

    # stores description for back conversion
    schema["description"] = root.xpath(
        "./xs:annotation/xs:documentation[2]/text()", namespaces=ns
    )[0]

    # store any schema detail items
    schemadetail = root.xpath(
        "./xs:annotation/xs:appinfo/lx:schemadetail", namespaces=ns
    )[0]
    for att in schemadetail.attrib:
        schema["schemadetail"][att] = schemadetail.attrib[att]

    # store any schema detail items
    schemacustomisation = root.xpath(
        "./xs:annotation/xs:appinfo/li:schemacustomisation", namespaces=ns
    )
    if len(schemacustomisation) > 0:
        schemacustomisation = schemacustomisation[0]
        schema["schemacustomisation"] = {}

        for customization_type in schemacustomisation:
            new_array = []
            for path in customization_type:
                new_array.append(path.text)

            schema["schemacustomisation"][
                customization_type.tag.replace(
                    "{lixi.org.au/schema/appinfo_instructions}", ""
                )
            ] = new_array

    # parses and converts all xs:simpleType for definitions
    for simple_type in root.findall("./xs:simpleType", namespaces=ns):
        add_simpleType(simple_type, schema, definitions)

    # parses and converts all xs:complexType for definitions
    for complex_type in root.findall("./xs:complexType", namespaces=ns):
        definitions[complex_type.attrib.get("name")] = add_element(
            complex_type, schema, definitions, True, definitions
        )

    # recurse and add all elemenets
    for element in root.findall("./xs:element", namespaces=ns):
        properties[element.attrib.get("name")] = add_element(
            element, schema, properties, False, definitions
        )

    return schema


def convert_to_xml_schema_string(jsonstring):
    """Converts a LIXI JSON schema to XML schema.
    
    Args: 
        jsonstring (:obj:`str`, required):  A json LIXI schema string.
    
    Returns:
        A XML LIXI schema string.

    """
    json_schema = json.loads(jsonstring)
    xsd_schema = convert_to_xml_schema(json_schema)

    return etree.tostring(xsd_schema, pretty_print=True).decode("utf-8")


def convert_to_xml_schema(json_schema):
    """Converts a LIXI JSON schema to XML schema.
    
    Args: 
        json_schema (:obj:`dict`, required): A json LIXI schema dict.
    
    Returns:
        A lxml etree LIXI schema.
    """

    simpleTypesLists = []
    simpleTypesPatterns = []
    simpleTypesTypes = []
    isAnnotated = False
    isCustom = False

    def append_elements_in_order(parent, new_element):
        siblings = parent.getchildren()
        if len(siblings) == 0:
            parent.append(new_element)
        else:
            found = False
            for sibling in siblings:
                if sibling.tag == "{http://www.w3.org/2001/XMLSchema}element":
                    if (
                        sibling.attrib["name"].lower()
                        > new_element.attrib["name"].lower()
                    ):
                        parent.insert((parent.index(sibling)), new_element)
                        found = True
                        break
            if not found:
                parent.append(new_element)

    def append_attributes_in_order(parent, new_element):
        siblings = parent.getchildren()
        if len(siblings) == 0:
            parent.append(new_element)
        else:
            found = False

            for sibling in siblings:
                if sibling.tag == "{http://www.w3.org/2001/XMLSchema}attribute":
                    if (
                        sibling.attrib["name"].lower()
                        > new_element.attrib["name"].lower()
                    ):
                        parent.insert((parent.index(sibling)), new_element)
                        found = True
                        break
            if not found:
                parent.append(new_element)

    def append_simpleTypes_in_order(orderlist, new_element):

        if len(orderlist) == 0:
            orderlist.append(new_element)
        else:
            found = False

            for sibling in orderlist:
                if sibling.tag == "{http://www.w3.org/2001/XMLSchema}simpleType":
                    if (
                        sibling.attrib["name"].lower()
                        > new_element.attrib["name"].lower()
                    ):
                        orderlist.insert((orderlist.index(sibling)), new_element)
                        found = True
                        break
            if not found:
                orderlist.append(new_element)

    def ns_element(namespace, name):
        return etree.Element(etree.QName(namespace, name), nsmap=ns)

    def create_schema_annotation(parent):

        isAnnotated = False
        isCustom = False

        doc1 = ns_element(xs, "documentation")
        doc1.text = "IMPORTANT: Copyright in this schema is owned by LIXI Limited and use of the schema is controlled by the LIXI End User Licence Agreement (EULA) which can be downloaded from http://www.lixi.org.au. The EULA does not permit you to extend the schema through the addition of new elements or attributes, or modify the content model or any data values in a manner that would be inconsistent with the Standard. However you are permitted to restrict the schema such that it specifies the structures for capturing only the information you wish to receive. This statement in no way limits or modifies the terms of the EULA, and you should ensure that you are familiar with your rights and obligations under the EULA before using the schema."

        doc2 = ns_element(xs, "documentation")
        doc2.text = json_schema["description"]

        appinfo = ns_element(xs, "appinfo")

        schemadetail = ns_element(lx, "schemadetail")

        for key, item in json_schema["schemadetail"].items():

            if key == "annotation":
                if item == "Full":
                    isAnnotated = True

            if key == "type":
                if item == "Custom":
                    isCustom = True

            schemadetail.set(key, item)

        if isCustom:
            schemacustomisation = ns_element(li, "schemacustomisation")

            for key, item in json_schema["schemacustomisation"].items():

                customization_type = ns_element(li, key)
                for path in item:
                    lxpath = ns_element(li, "path")
                    lxpath.text = path
                    customization_type.append(lxpath)
                schemacustomisation.append(customization_type)

            appinfo.append(schemacustomisation)

        appinfo.append(schemadetail)
        annotation = ns_element(xs, "annotation")
        annotation.append(doc1)
        annotation.append(doc2)
        annotation.append(appinfo)

        parent.append(annotation)

        return isAnnotated, isCustom

    def create_annotation(parent, json_element):

        if isAnnotated:

            annotation = ns_element(xs, "annotation")

            documentation = ns_element(xs, "documentation")
            documentation.text = json_element["description"]

            appinfo = ns_element(xs, "appinfo")

            lx_path = ns_element(lx, "path")
            lx_path.text = json_element["path"]

            lx_label = ns_element(lx, "label")
            lx_label.text = json_element["label"]

            if isCustom:
                if "custom_description" in json_element:
                    lx_CustomDocumentation = ns_element(lx, "CustomDocumentation")
                    lx_CustomDocumentation.text = json_element["custom_description"]
                    appinfo.append(lx_CustomDocumentation)

                if "custom_excerpt" in json_element:
                    lx_CustomExcerpt = ns_element(lx, "CustomExcerpt")
                    lx_CustomExcerpt.text = json_element["custom_excerpt"]
                    appinfo.append(lx_CustomExcerpt)
            
            appinfo.append(lx_path)
            appinfo.append(lx_label)            
            
            if "references" in json_element:
                lx_references = ns_element(lx, "references")
                for target in json_element["references"]:
                    lx_target = ns_element(lx, "target")
                    lx_target.text = target
                    lx_references.append(lx_target)
                appinfo.append(lx_references)

            annotation.append(documentation)
            annotation.append(appinfo)

            parent.append(annotation)

    def create_element(parent, json_element, name=None, use=None):

        element = ns_element(xs, "element")

        # name
        if name != None:
            element.set("name", name)

        # special cases like inline attachement
        if "properties" in json_element:
            if "$" in json_element["properties"]:
                xml_type = str(json_element["properties"]["$"]["$ref"]).replace(
                    "#/definitions/", ""
                )
                element.set("type", xml_type)

        # Type
        xml_type = None
        if "type" in json_element:
            if json_element["type"] == "object":
                json_type = json_element["type"]

                if json_type == "base64Binary":
                    xml_type = "base64BinaryType"
                elif json_type == "base64Binary":
                    xml_type = "currencyType"
                elif json_type == "dateTime":
                    xml_type = "dateTimeType"
                elif json_type == "date":
                    xml_type = "dateType"
                elif json_type == "decimal":
                    xml_type = "decimalType"
                elif json_type == "gYear":
                    xml_type = "gYearType"
                elif json_type == "integer":
                    xml_type = "integerType"
                elif json_type == "ID":
                    xml_type = "percentType"
                elif json_type == "IDREF":
                    xml_type = "referenceType"
                elif json_type == "string":
                    xml_type = "stringType"
                elif json_type == "ID":
                    xml_type = "uniqueIDType"
            elif json_element["type"] == "array":
                if "$ref" in json_element["items"]:
                    xml_type = str(json_element["items"]["$ref"]).replace(
                        "#/definitions/", ""
                    )

        elif "$ref" in json_element:
            xml_type = str(json_element["$ref"]).replace("#/definitions/", "")

        if xml_type != None:
            element.set("type", xml_type)

        if parent.tag != "{http://www.w3.org/2001/XMLSchema}schema":
            # Type
            min_occurs = "0"  # default LIXI preference
            max_occurs = "1"

            if use == "required":
                min_occurs = "1"

            if "type" in json_element:
                if json_element["type"] == "array":

                    min_occurs = "0"
                    max_occurs = "unbounded"

                    if "minItems" in json_element:
                        min_occurs = str(json_element["minItems"])

                    if "maxItems" in json_element:
                        max_occurs = str(json_element["maxItems"])

            element.set("minOccurs", min_occurs)
            element.set("maxOccurs", max_occurs)

        create_annotation(element, json_element)

        create_complexType(element, json_element, name=None)

        append_elements_in_order(parent, element)

    def create_attribute(parent, json_element, name=None, use=None):
        attribute = ns_element(xs, "attribute")

        # name
        attribute.set("name", name[1:])

        # Type
        xml_type = None
        if "type" in json_element:
            xml_type = "stringType"
            json_type = json_element["type"]

            if json_type == "base64Binary":
                xml_type = "base64BinaryType"
            elif json_type == "dateTime":
                xml_type = "dateTimeType"
            elif json_type == "date":
                xml_type = "dateType"
            elif json_type == "decimal":
                xml_type = "decimalType"
            elif json_type == "gYear":
                xml_type = "gYearType"
            elif json_type == "integer":
                xml_type = "integerType"
            elif json_type == "IDREF":
                xml_type = "referenceType"
            elif json_type == "string":
                xml_type = "stringType"
            elif json_type == "ID":
                xml_type = "uniqueIDType"

        elif "$ref" in json_element:
            xml_type = str(json_element["$ref"]).replace("#/definitions/", "")

        if xml_type != None:
            attribute.set("type", xml_type)

        # use
        attribute.set("use", use)

        # fixed
        if xml_type == "stringType" and "enum" in json_element:
            if len(json_element["enum"]) == 1:
                attribute.set("fixed", json_element["enum"][0])

        # annotation
        create_annotation(attribute, json_element)

        append_attributes_in_order(parent, attribute)

    def create_complexType(parent, json_element, name=None):

        complexType = ns_element(xs, "complexType")

        if name != None:
            complexType.set("name", name)
            create_annotation(complexType, json_element)

        if "properties" in json_element:  ##type is object

            ##run loop twice to account for sequence
            sequence = ns_element(xs, "sequence")
            for key, item in json_element["properties"].items():
                if key[:1] != "@" and key != "$" and key != "oneOf":
                    if "required" in json_element:
                        use = (
                            "required"
                            if key in json_element["required"]
                            else "optional"
                        )
                    else:
                        use = "optional"

                    create_element(sequence, item, key, use)

            if len(sequence.getchildren()) > 0:
                complexType.append(sequence)
            
                
            for key, item in json_element["properties"].items():
                
                    
                if key == "oneOf":
                    
                    choice = ns_element(xs, "choice")
        
                    # for choice_item in json_element["properties"]["oneOf"]:
                    for choice_item in json_element["oneOf"]:
                        for key, item in choice_item["properties"].items():
                            if key[:1] != "@":
                                if "required" in choice_item:
                                    use = (
                                        "required"
                                        if key in choice_item["required"]
                                        else "optional"
                                    )
                                else:
                                    use = "optional"
        
                                create_element(choice, item, key, use)
        
                    complexType.append(choice)
                    
            for key, item in json_element["properties"].items():       
                
                if key[:1] == "@":

                    if "required" in json_element:
                        use = (
                            "required"
                            if key in json_element["required"]
                            else "optional"
                        )
                    else:
                        use = "optional"

                    create_attribute(complexType, item, key, use)                    
                    

        elif "items" in json_element:  ##type is array:

            ##run loop twice to account for sequence
            sequence = ns_element(xs, "sequence")
            if "properties" in json_element["items"]:
                for key, item in json_element["items"]["properties"].items():
                    if key[:1] != "@" and key != "$":
                        if "required" in json_element:
                            use = (
                                "required"
                                if key in json_element["required"]
                                else "optional"
                            )
                        else:
                            use = "optional"

                        create_element(sequence, item, key, use)

            if len(sequence.getchildren()) > 0:
                complexType.append(sequence)

            if "properties" in json_element["items"]:
                for key, item in json_element["items"]["properties"].items():
                    if key[:1] == "@":
                        if "required" in json_element["items"]:
                            use = (
                                "required"
                                if key in json_element["items"]["required"]
                                else "optional"
                            )
                        else:
                            use = "optional"

                        create_attribute(complexType, item, key, use)

        elif "oneOf" in json_element:  ##type is choice/oneof:

            choice = ns_element(xs, "choice")

            for choice_item in json_element["oneOf"]:

                for key, item in choice_item["properties"].items():
                    if key[:1] != "@":
                        if "required" in choice_item:
                            use = (
                                "required"
                                if key in choice_item["required"]
                                else "optional"
                            )
                        else:
                            use = "optional"

                        create_element(choice, item, key, use)

            complexType.append(choice)

            first_choice_item = json_element["oneOf"][
                0
            ]  ##if there are attributes for choice they added to all oneOf items in json hence grabbing attributes from the first only will suffice

            for key, item in first_choice_item["properties"].items():
                if key[:1] == "@":
                    if "required" in json_element:
                        use = (
                            "required"
                            if key in json_element["required"]
                            else "optional"
                        )
                    else:
                        use = "optional"

                    create_attribute(complexType, item, key, use)

        if len(complexType.getchildren()) > 0:
            parent.append(complexType)

    def create_enumeration(parent, json_element, name=None):

        enumeration = ns_element(xs, "enumeration")

        enumeration.set("value", name)
        if isAnnotated:
            create_annotation(enumeration, json_element["enum_definitions"][name])

        parent.append(enumeration)

    def create_simpleType(parent, json_element, name=None):
        simpleType = ns_element(xs, "simpleType")

        if name != None:
            simpleType.set("name", name)
            create_annotation(simpleType, json_element)

        restriction = ns_element(xs, "restriction")

        if "enum" in json_element:  ##simpletype is a List
            restriction.set("base", "xs:token")

            for enum in json_element["enum"]:
                create_enumeration(restriction, json_element, enum)

            simpleType.append(restriction)
            append_simpleTypes_in_order(simpleTypesLists, simpleType)

        elif "pattern" in json_element:  ##simpletype is Pattern

            patt = str(json_element["pattern"]).replace("\\\\", "\\")[1:-1]

            restriction.set("base", "xs:string")
            pattern = ns_element(xs, "pattern")
            pattern.set("value", patt)
            restriction.append(pattern)
            simpleType.append(restriction)
            append_simpleTypes_in_order(simpleTypesPatterns, simpleType)

        elif "$ref" in json_element:  ##simpletype is a Type

            if name == "base64BinaryType":
                restriction.set("base", "xs:base64Binary")
            elif name == "currencyType":
                restriction.set("base", "xs:decimal")
            elif name == "dateTimeType":
                restriction.set("base", "xs:dateTime")
            elif name == "dateType":
                restriction.set("base", "xs:date")
            elif name == "decimalType":
                restriction.set("base", "xs:decimal")
            elif name == "gYearType":
                restriction.set("base", "xs:gYear")
            elif name == "integerType":
                restriction.set("base", "xs:integer")
            elif name == "percentType":
                restriction.set("base", "xs:decimal")
            elif name == "referenceType":
                restriction.set("base", "xs:IDREF")
            elif name == "stringType":
                restriction.set("base", "xs:string")
            elif name == "uniqueIDType":
                restriction.set("base", "xs:ID")

            simpleType.append(restriction)
            append_simpleTypes_in_order(simpleTypesTypes, simpleType)

        elif (
            "minimum" in json_element or "maximum" in json_element
        ):  ##simpletype is a Range

            if json_element["type"] == "decimal":
                restriction.set("base", "xs:decimal")
            elif json_element["type"] == "integer":
                restriction.set("base", "xs:integer")
            elif json_element["type"] == "string":
                if "format" in json_element:
                    if json_element["format"] == "lixi-year":
                        restriction.set("base", "xs:gYear")
                    elif json_element["format"] == "lixi-datetime":
                        restriction.set("base", "xs:dateTime")
                    elif json_element["format"] == "lixi-date":
                        restriction.set("base", "xs:date")
                else:
                    restriction.set("base", "xs:string")

            if "minimum" in json_element:
                minInclusive = ns_element(xs, "minInclusive")
                minInclusive.set("value", json_element["minimum"])
                restriction.append(minInclusive)

            if "maximum" in json_element:
                maxInclusive = ns_element(xs, "maxInclusive")
                maxInclusive.set("value", json_element["maximum"])
                restriction.append(maxInclusive)

            simpleType.append(restriction)
            append_simpleTypes_in_order(simpleTypesTypes, simpleType)

    json_schema = json.loads(json_schema)
    isAnnotated = False

    Schema = ns_element(xs, "schema")
    Schema.set("elementFormDefault", "qualified")
    Schema.set("attributeFormDefault", "unqualified")
    Schema.set("version", "1.0")
    isAnnotated, isCustom = create_schema_annotation(Schema)

    # Package
    create_element(Schema, json_schema["properties"]["Package"], "Package", "required")

    # cater to all complex types and simple types
    for key, item in json_schema["definitions"].items():
        if "type" in item:
            if item["type"] == "object":
                create_complexType(Schema, item, key)
            elif key.endswith("Pattern"):
                create_simpleType(Schema, item, key)
            elif key.endswith("Range"):
                create_simpleType(Schema, item, key)
        else:
            if key.endswith("Type"):
                create_simpleType(Schema, item, key)
            elif key.endswith("List"):
                create_simpleType(Schema, item, key)

    # Sort the xml schema as per lixi order.
    simpleTypeList = simpleTypesLists + simpleTypesPatterns + simpleTypesTypes
    for simpletype in simpleTypeList:
        Schema.append(simpletype)

    return Schema


### Output code
# with io.open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_19-Annotated.xsd','r', encoding="utf-8") as xml_file:
# xml_schema = xml_file.read()
# parser = etree.XMLParser(remove_blank_text=True)
# xsd_schema = etree.fromstring(xml_schema, parser)
# json_schema = convert_to_json_schema(xsd_schema)
# with io.open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_19_RFC-Annotated-CREATED.json',"wb+") as f:
# string = json.dumps(json_schema, sort_keys=True, indent=4, ensure_ascii=False).encode('utf-8')
# f.write(string)
###
# with io.open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_19_RFC-Annotated-CREATED.json','r', encoding="utf-8") as json_file:
# json_schema = json_file.read()
# xml_schema = convert_to_xml_schema(json_schema)
# with io.open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_19-Annotated-CREATED.xsd',"wb+") as f:
# string = etree.tostring(xml_schema, pretty_print=True, encoding='UTF-8')
# f.write(string)


##Code for better diff comparison
# xsd_schema = etree.parse('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_21-Annotated.xsd')
# json_schema = convert_to_json_schema(xsd_schema.getroot())
# with open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_21-Annotated-CREATED.json',"w+") as f:
# f.write(json.dumps(json_schema, sort_keys=True, indent=4))


# with open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_21-Annotated-CREATED.json','r') as json_file:
# json_schema = json_file.read()
# xml_schema = convert_to_xml_schema(json_schema)
# with open('C:/Users/compb/Downloads/xml/LIXI-CAL-2_6_21-Annotated-CREATED.xsd',"w+") as f:
# f.write(etree.tostring(xml_schema, pretty_print=True).decode('utf-8'))

# print("Done")



from lxml import etree
import re, time

if __name__ == "_LIXI":
    import _explode_schema
else:
    from lixi import (
        _explode_schema
    )

ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}


def get_paths_for_elements(message=None, paths_list=None, schema=None):
    """
    Obtains all the element paths from a LIXI message or a paths list. In case of a paths list, obtains all the correct paths and referenced type paths.

    Args:
        message (:obj:`lixi._Message`, optional): A LIXI message instance.
        paths_list (:obj:`list`, optional): A Python list of paths.
        schema (:obj:`lxml.etree`, required): A etree object of a LIXI XML Annotated Schema.

    Returns:
        A Python list of paths that need to be included for customization. 
    """

    start = time.time()
    paths = []

    # Determine source
    if message != None:
        message_xml = message.xml_package
        message_tree = etree.ElementTree(message_xml)

        for e in message_tree.iter():
            # Get the base path of element, get it to lixi path
            element_path = message_tree.getpath(e)  ##Gets the path in lxml format
            element_path = element_path[1:]  ##Removes the first '/' from path
            element_path = element_path.replace(
                "/", "."
            )  ##Replaces '/' with '.' for LIXI Paths
            element_path = "".join(
                [i for i in element_path if not i.isdigit() and i != "[" and i != "]"]
            )  ##Removes numbers, and [] characters

            # Add base path
            if re.match(r"^[a-zA-Z0-9_@\.]*$", element_path):
                paths.append(element_path)

            # Check for elements attributes
            for att in e.attrib:
                att_path = element_path + "." + att

                if re.match(r"^[a-zA-Z0-9_@\.]*$", att_path):
                    paths.append(att_path)

    elif paths_list != None:
        # ensure element path is entered in the correct format
        for path in paths_list:
            schema_path_parent_breaks = path.split(".")
            append_path = ""

            for path_break in schema_path_parent_breaks:
                append_path = (
                    append_path + "." + path_break if append_path != "" else path_break
                )

                if append_path not in paths:
                    paths.append(append_path)

    paths.sort()
    # print(paths)
    final_paths = []
    complextype_paths = []
    simpletype_paths = []
    current_complex_type = None
    parent_path = None

    # dict of paths for effeciency, so lx:path xpath can be avoided
    dict_path_elem = {}
    all_schema_path_elements = schema.xpath("//lx:path", namespaces=ns)
    for path_elem in all_schema_path_elements:
        dict_path_elem[path_elem.text] = path_elem

    # unravel all complexTypes
    for message_path in paths:

        if re.match(r"^[a-zA-Z0-9_@\.]*$", message_path):

            if message_path not in dict_path_elem:

                # A case when exploded complex paths come which are non searchable, so they have to be broken into something that can be searched in an unexploded schema
                orig_message_path = message_path
                #print(orig_message_path)
                cut_message_path = message_path.replace(parent_path, "")
                message_path = current_complex_type + cut_message_path

                ##complex within a complex
                if message_path not in dict_path_elem:

                    paths = cut_message_path[1:].split(".")
                    index = 0
                    is_complex = False
                    new_complex_type = current_complex_type
                    found_complex = None

                    while is_complex == False:
                        new_complex_type = new_complex_type + "." + paths[index]

                        find_it = schema.xpath(
                            '//xs:element/xs:annotation/xs:appinfo/lx:path[text()="'
                            + new_complex_type
                            + '"]',
                            namespaces=ns,
                        )
                        if len(find_it) > 0:
                            if (
                                "type"
                                in find_it[0].getparent().getparent().getparent().attrib
                            ):
                                is_complex = True
                                found_complex = (
                                    find_it[0]
                                    .getparent()
                                    .getparent()
                                    .getparent()
                                    .attrib["type"]
                                )
                            else:
                                index = index + 1
                        else:
                            index = index + 1

                    if found_complex != None:

                        remaining_path = message_path.replace(new_complex_type, "")

                        if found_complex not in complextype_paths:
                            complextype_paths.append(found_complex)

                        message_path = found_complex + remaining_path

                searched_simple_path = dict_path_elem[message_path]
                searched_simple_parent = (
                    searched_simple_path.getparent().getparent().getparent()
                )

                if "type" in searched_simple_parent.attrib:
                    typename = searched_simple_parent.attrib["type"]

                    if typename in dict_path_elem:
                        if typename not in simpletype_paths:
                            simpletype_paths.append(typename)

                if message_path not in complextype_paths:
                    complextype_paths.append(message_path)

            else:

                # case when path is searchable
                parent = (
                    dict_path_elem[message_path].getparent().getparent().getparent()
                )

                if "type" in parent.attrib:
                    typename = parent.attrib["type"]

                    if typename in dict_path_elem:
                        searched_path = (
                            dict_path_elem[typename].getparent().getparent().getparent()
                        )
                        tag = searched_path.tag

                        if tag == "{http://www.w3.org/2001/XMLSchema}complexType":
                            current_complex_type = typename
                            parent_path = message_path

                            if typename not in complextype_paths:
                                complextype_paths.append(typename)

                        elif tag == "{http://www.w3.org/2001/XMLSchema}simpleType":

                            if typename not in simpletype_paths:
                                simpletype_paths.append(typename)

                final_paths.append(message_path)

    # simpleTypes
    for simple_type in simpletype_paths:
        last_four = simple_type[-4:]

        if last_four == "List":
            searched_simple_path = schema.xpath(
                './xs:simpleType[@name="' + simple_type + '"]', namespaces=ns
            )
            simpletype_paths = simpletype_paths + searched_simple_path[0].xpath(
                "./xs:restriction/xs:enumeration/xs:annotation/xs:appinfo/lx:path/text()",
                namespaces=ns,
            )

    final_paths.sort()
    complextype_paths.sort()
    simpletype_paths.sort()

    final_paths = final_paths + complextype_paths + simpletype_paths
    final_paths = list(dict.fromkeys(final_paths))
    return final_paths

def get_exploded_paths_for_schema_elements(annotated_schema):
    """
    Gets all the ecploded element paths from a LIXI XML schema.

    Args:
        annotated_schema (:obj:`lxml.etree`, required): A etree object of a LIXI XML Annotated Schema.
            
    Returns:
        A Python list of paths in the LIXI Schema. 
    """
    
    exploded_schema = _explode_schema.main(annotated_schema)
    
    return exploded_schema.xpath("//lx:path/text()", namespaces=ns)
    
    ## dict of paths for effeciency, so lx:path xpath can be avoided
    #package = annotated_schema.xpath("./xs:element", namespaces=ns)
    #complex_types = annotated_schema.xpath("./xs:complexType//lx:path/text()", namespaces=ns)
    #simple_types = annotated_schema.xpath("./xs:simpleType//lx:path/text()", namespaces=ns)
    
    #element_paths = package[0].xpath("//xs:element/xs:annotation/xs:appinfo/lx:path", namespaces=ns)
    #attribute_paths = package[0].xpath("//xs:attribute/xs:annotation/xs:appinfo/lx:path/text()", namespaces=ns)
    
    #all_element_paths = []
    #for lxpath in element_paths:
        
        #if 'type' in lxpath.getparent().getparent().getparent().attrib:
            
            #complextype = lxpath.getparent().getparent().getparent().attrib['type']
            #if complextype in complex_types:
                
                #parent_path = lxpath.text
                #for complex_path in complex_types:
                    #if complextype in complex_path:
                        #all_element_paths.append(complex_path.replace(complextype, parent_path))
                
            #else:
                #all_element_paths.append(lxpath.text)
        #else:
            #all_element_paths.append(lxpath.text)
                
    
    #final_paths = all_element_paths + attribute_paths + simple_types
    #final_paths.sort()
    
    ##print(json.dumps(final_paths, indent=3))
    
    #return final_paths

def get_paths_for_schema_elements(annotated_schema):
    """
    Gets all the element paths from a LIXI XML schema.

    Args:
        annotated_schema (:obj:`lxml.etree`, required): A etree object of a LIXI XML Annotated Schema.
            
    Returns:
        A Python list of paths in the LIXI Schema. 
    """
    # Xpath to get all schema element paths
    return annotated_schema.xpath("//lx:path/text()", namespaces=ns)


def get_blacklist_paths_for_customization(
    transaction_type, message_paths, schema_paths
):
    """
    Generates a string of blacklist customization restrictions xml for a customized LIXI schema.

    Args:
        transaction_type (:obj:`str`, required): The transaction type of the LIXI schema.
        message_paths (:obj:`lxml.etree`, required): All element paths from a LIXI message.
        schema_paths (:obj:`lxml.etree`, required): All element paths from a LIXI schema.
            
    Returns:
        A string of blacklist customization restrictions xml that can be utilized by this library to generate a custom schema.  
    """

    # Get list of paths to remove:
    paths_to_remove = set(schema_paths) - set(message_paths)
    paths_to_remove = list(paths_to_remove)

    paths_to_remove.sort()

    # Form a string of customization instructions.
    customization_instructions = ""
    customization_instructions += (
        '''<Customisations xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="LIXI-Customisation-By-Restriction.xsd" LIXICode="LIXILIXI" CustomisationCode="v1" LIXITransactionType="'''
        + transaction_type
        + """">\n"""
    )

    # Some paths that can't be removed as they are reserved.
    paths_that_cant_removed = [
        "Package",
        "Package.SchemaVersion",
        "Package.SchemaVersion.LIXICustomVersion",
        "Package.SchemaVersion.LIXITransactionType",
        "Package.SchemaVersion.LIXIVersion",
        "Package.Publisher",
        "Package.Publisher.LIXICode",
        "Package.Recipient",
        "Package.Recipient.LIXICode",
        "Package.Instructions",
        "Package.Instructions.AccountCreationInstructions",
        "Package.Instructions.AccountCreationInstructions.Submit",
        "Package.Instructions.ApplicationInstructions",
        "Package.Instructions.ApplicationInstructions.Submit",
        "Package.Instructions.CreditDecisionInstructions",
        "Package.Instructions.CreditDecisionInstructions.Submit",
        "Package.Instructions.DocumentsAndSettlementInstructions",
        "Package.Instructions.DocumentsAndSettlementInstructions.Submit",
        "Package.Instructions.LendersMortgageInsuranceInstructions",
        "Package.Instructions.LendersMortgageInsuranceInstructions.Submit",
        "Package.Instructions.ServiceabilityInstructions",
        "Package.Instructions.ServiceabilityInstructions.Submit",
        "Package.Instructions.ValuationInstructions",
        "Package.Instructions.ValuationInstructions.Submit",
    ]
    
    
    #If parent removed then no need to add child.
    redo_list = []
    for path in paths_to_remove:
        if path not in paths_that_cant_removed:
            
            if '.' in path:
                
                parent_in = False
                
                splits = path.split('.')
                
                i=0
                check_path = splits[i]
                while check_path != path:
                    
                    if check_path in redo_list:
                        parent_in = True
                    
                    i = i + 1
                    check_path = check_path + '.'+ splits[i]
                
                if parent_in == False:
                    redo_list.append(path)
            else:
                redo_list.append(path)
    
    #If simpletype/complextype are removed from base then no need to add that in instructions as it clutters it and custom schema already handles it anyhow.
    redo_redo_list = []
    for path in redo_list:
        if 'Package' in path or '.' in path:
            redo_redo_list.append(path)
    
    
    for path in redo_redo_list:

        if path not in paths_that_cant_removed:
            customization_instructions += (
                '<CustomiseItem Exclude="Yes"><Path>'
                + path
                + "</Path></CustomiseItem>\n"
            )

    customization_instructions += "</Customisations>"
    #with open('./simp/ins.xml','w+') as f:
        #f.write(customization_instructions)
        
    return customization_instructions



import os as _os

if 'test_version' in _os.environ:
    __version__ = str(_os.environ.get('test_version'))
else:
    __version__ = 'test' 


__all__ = [
    #"set_credentials",
    "set_schema_folder",
    "get_schema_paths",
    "get_custom_schema",
    "get_custom_schema_from_message_folder",
    "get_json_schema",
    "get_xml_schema",
    "get_schema_documentation",
    "read_message",
]


if __name__ == "__main__":
    from _LIXI import (
        LIXI as _LIXI,
        LIXIValidationError
    )
    from _Message import Message as _Message
else:
    from lixi._LIXI import (
        LIXI as _LIXI,
        LIXIValidationError
    )
    from lixi._Message import Message as _Message


###################
## ENTRY POINTS
###################


#def set_credentials(access, secret):
    #"""Sets the secret and access keys for a LIXI member, which would fetch a schema from the LIXI online repository. Used before doing any operations that require a schema. Both access key and secret key can be requested from LIXI admin. These are unique for an organization/member.  
    
    #Args:
        #access (:obj:`str`, required): A LIXI member's assigned access key. 
        #secret (:obj:`str`, required): A LIXI member's assigned secret key. 
    
    #Raises:
        #LIXIInvalidSyntax: If 'access' is not specified or is not str. 
        #LIXIInvalidSyntax: If 'secret' is not specified or is not str. 
    #"""

    #_LIXI.getInstance().set_credentials(access, secret)


def set_schema_folder(folder_path, create_config=True, force_update=False, lambda_check=False):
    """Sets a directory path to fetch LIXI schemas from. Used before doing any operations that require a schema. Allows schema read from folder and a re-read of folder in case a file was added on runtime.   
    
    Args:
        folder_path (:obj:`str`, required): A path to a folder containing LIXI schema files.
        create_config (:obj:`bool`, required): A flag to indicate whether a folder read is required. Defaults to True.
        force_update (:obj:`bool`, optional): A flag to force update of the folder read. Defaults to False.
        lambda_check (:obj:`bool`, optional): A flag to indicate library being run on lambda. Defaults to False.

    Raises: 
        LIXIInvalidSyntax: If folder_path is not specified.  
        LIXIResouceNotFoundError: If folder_path provided does not exist. 
    """
    _LIXI.getInstance().set_schema_path(folder_path, create_config, force_update, lambda_check)


def get_schema_paths(
    lixi_transaction_type=None,
    lixi_version=None,
    custom_version=None,
    schema_string=None,
    schema_path=None,
    exploded = False
):
    """Fetches all of the elements paths of the LIXI schema provided as a Python List of paths. 

    Args:
        lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None. 
        lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided. 
        custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None. 
        schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None. 
        schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.
        exploded (:obj:`bool`, optional): Variable to indicate if the returned schema should be exploded or not. 

    Returns:
        A python list of all paths in the LIXI schema. 

    Raises:
        LIXIInvalidSyntax: Errors if inputs are not in the correct format. 
        LIXIResouceNotFoundError: Errors if inputs are not found. 
        LIXIValidationError: Validation errors for inputs. 
    """
    return _LIXI.getInstance().get_schema_paths(
        lixi_transaction_type,
        lixi_version,
        "xml",
        custom_version,
        schema_string,
        schema_path,
        exploded
    )


def get_custom_schema(
    instructions_string=None,
    instructions_path=None,
    elements_array=None,
    lixi_transaction_type=None,
    lixi_version=None,
    schema_string=None,
    schema_path=None,
    output_name=None,
    output_folder=None,
    output_type="xml",
    output_instructions=False
):
    """Generates a custom LIXI schema based on the given customization instructions source. This produced schema is based on the project (https://standards.lixi.org.au/lixi2/CustomisationByRestriction) that derives a subschema that only uses sub set of elements available in the full LIXI schema. This produced schema is a valid LIXI schema. 

    Args:
        instructions_string (:obj:`str`, optional): Customization instructions provided as a string. Defaults to None. 
        instructions_file (:obj:`str`, optional): Customization instructions provided as a path. Defaults to None. 
        elements_array (:obj:`str`, optional): Customization paths to keep in an array. Defaults to None.
        lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None. 
        lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided. 
        schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None. 
        schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None. 
        output_name (:obj:`str`, optional): String to set as a name of the custom schema. Defaults to None. 
        output_folder (:obj:`str`, optional): Path to write the produced lixi schema to. Defaults to None. 
        output_type (:obj:`str`, optional): The output type of the custom LIXI schema produced. Currently only accepts 'xml' and 'json'. Defaults to 'xml'.
        output_instructions (:obj:`boolean`, optional): A flag to indicate whether a customization instructions need to be saved to a path. Defaults to False. 

    Returns: 
        A customised LIXI schema as a string which can also be to the output folder specified. 

    Raises: 
        LIXIInvalidSyntax: Validation errors for the lixi version or transaction type. 
        LIXIResouceNotFoundError: If path provided does not exist. 
        LIXIValidationError: validation error in the customization instructions. 
    """

    return _LIXI.getInstance().generate_custom_schema(
        instructions_string,
        instructions_path,
        elements_array,
        lixi_transaction_type,
        lixi_version,
        schema_string,
        schema_path,
        output_name,
        output_folder,
        output_type,
        output_instructions
    )

def get_custom_schema_from_message_folder(
    message_folder=None,
    lixi_transaction_type=None,
    lixi_version=None,
    schema_path=None,    
    output_name=None,
    output_folder=None,
    output_type="xml"
):
    """Generates a custom LIXI schema based on the given customization instructions source. 

    Args:
        message_folder (:obj:`str`, required): A message folder to generate the customization instructions from. Defaults to None.
        lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None. 
        lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched for customization. Should be in the format of '2.6.24'. Defaults to the latest version if only lixi_transaction_type is provided. 
        schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None. 
        output_name (:obj:`str`, optional): String to set as a name of the custom schema. Defaults to None. 
        output_folder (:obj:`str`, optional): Path to write the produced lixi schema to. Defaults to None. 
        output_type (:obj:`str`, optional): The output type of the custom LIXI schema produced. Currently only accepts 'xml' and 'json'. Defaults to 'xml'. 

    Returns: 
        A customised LIXI schema as a string which can also be to the output folder specified.  

    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path. 
        LIXIInvalidSyntax: If the schema file is not well formed. 
        LIXIValidationError: If there are validation errors in the message file. 
    """
    
    instructions = ''
    message_items = []
    
    # Read message folder
    for message_name in _os.listdir(message_folder):
        
        full_message_path = _os.path.join(message_folder, message_name)
        
        if _os.path.isfile(full_message_path) and ('.xml' in message_name or '.json' in message_name):
            
            try:
                # Read individual message
                lixi_message = _Message(None, full_message_path, None)

                # Validate the Message
                isvalid, result, error_log = lixi_message.validate()
                if isvalid == False:
                    raise LIXIValidationError(result, lixi_message, error_log)
                
                # Generate message paths from Message
                message_items = message_items + lixi_message.get_message_paths()
            
            except Exception as e:
                raise Exception('Error in message, ' +message_name+'\n'+ str(e))
     
    if message_items != '':
        message_items = list(set(message_items))
        message_items.sort()
        elements_array = message_items

        
        return _LIXI.getInstance().generate_custom_schema(
            None,
            None,
            elements_array,
            lixi_transaction_type,
            lixi_version,
            None,
            schema_path,
            output_name,
            output_folder,
            output_type,
        )        
    
    else:
        raise Exception('Custom schema not generated. ')

def get_json_schema(
    lixi_transaction_type=None,
    lixi_version=None,
    custom_version=None,
    schema_string=None,
    schema_path=None,
    output_path=None,
    return_annotated=False,
    exploded = False
):
    """Fetches a LIXI JSON schema dict object or converts from an LIXI XML schema. 

    Args:
        lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None. 
        lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided. 
        custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None. 
        schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None. 
        schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None. 
        output_path (:obj:`str`, optional): Path to write the fetched lixi schema to. Defaults to None.
        return_annotated (:obj:`boolean`, optional): Variable to indicate if the returned schema should be annotated or not.
        exploded (:obj:`boolean`, optional): Variable to indicate if the returned schema should be exploded or not.

    Result: 
        A string object of the LIXI JSON schema or saved to the output folder specified.  

    Raises:
        LIXIInvalidSyntax: Incorrect schema string/file fetched. 
        LIXIValidationError: Validation errors in the schema string/file fetched. 
        LIXIResouceNotFoundError: File not found or can't be written. 
    """

    return _LIXI.getInstance().fetch_json_schema(
        lixi_transaction_type,
        lixi_version,
        custom_version,
        schema_string,
        schema_path,
        output_path,
        return_annotated,
        exploded
    )


def get_xml_schema(
    lixi_transaction_type=None,
    lixi_version=None,
    custom_version=None,
    schema_string=None,
    schema_path=None,
    output_path=None,
    return_annotated=False,
    exploded=False
):
    """Fetches a LIXI XML schema Etree object or converts from an LIXI JSON schema.
    
    Args:  
        lixi_transaction_type (:obj:`str`, optional): The transaction type of the LIXI schema to be fetched. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.   
        lixi_version (:obj:`str`, optional): The version of the LIXI schema to be fetched. Should be in the format of '2.6.24'. Defaults to None. Defaults to the latest version if only lixi_transaction_type is provided.   
        custom_version (:obj:`str`, optional): The version of the LIXI custom schema to be fetched. Usually a complete file name of the custom schema. Defaults to None.   
        schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None.    
        schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None.    
        output_path (:obj:`str`, optional): Path to write the fetched lixi schema to. Defaults to None.
        return_annotated (:obj:`boolean`, optional): Variable to indicated if the returned schema should be annotated or not.     
        exploded (:obj:`boolean`, optional): Variable to indicate if the returned schema should be exploded or not.

    Result:   
        A string object of the LIXI XML schema or saved to the output folder specified.    

    Raises:   
        LIXIInvalidSyntax: Incorrect schema string/file fetched.    
        LIXIValidationError: Validation errors in the schema string/file fetched.    
        LIXIResouceNotFoundError: File not found or can't be written.    
    """

    return _LIXI.getInstance().fetch_xml_schema(
        lixi_transaction_type,
        lixi_version,
        custom_version,
        schema_string,
        schema_path,
        output_path,
        return_annotated,
        exploded
    )

def get_schema_documentation(
    schema_string, 
    output_folder=None, 
    output_name=None,
    glossary_string = None
):
    """Generates a custom LIXI schema based on the given customization instructions source. This produced schema is based on the project (https://standards.lixi.org.au/lixi2/CustomisationByRestriction) that derives a subschema that only uses sub-set of elements available in the full LIXI schema. This produced schema is a valid LIXI schema.

    Args:
        schema_string (:obj:`str`, required): LIXI schema provided as a string.
        output_folder (:obj:`str`, required): Path to write the produced lixi schema documentation to. If output_path is specified as None, a zip bytes file is returned.
        output_name (:obj:`str`, optional): Name of the sub-folder that contains the documentation. Defaults to None.      
        glossary_string (:obj:`str`, optional): LIXI Glossary update provided as a string. Defaults to None.
    
    Returns:
        Outputs the custom documentation for a schema to the specified folder. 
    
    Raises:
        LIXIInvalidSyntax: If the provided schema_string is not a string or is unable to be read.
        LIXIResouceNotFoundError: If path provided does not exist.
    """

    return _LIXI.getInstance().generate_schema_documentation(
        schema_string,
        output_folder,
        output_name,
        glossary_string
    )

def read_message(
    message=None, message_path=None, file_type=None, schema_text=None, schema_path=None
):
    """Reads a LIXI message XML. 

    Args:
        message (:obj:`str`, optional): A LIXI Message provided as a string. Defaults to None. 
        message_path (:obj:`str`, optional) : A LIXI Message provided as a path. Defaults to None. 
        file_type (:obj:`str`, optional): The type of the input LIXI Message given. Defaults to 'xml'. DEPRECATED. But have kept it for later use if needed. 
        schema_string (:obj:`str`, optional): LIXI schema provided as a string. Defaults to None. 
        schema_path (:obj:`str`, optional): LIXI schema provided as a path. Defaults to None. 

    Returns: 
        A LIXI message instance. 

    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path. 
        LIXIInvalidSyntax: If the schema file is not well formed. 
        LIXIValidationError: If there are validation errors in the message file. 
    """

    # Read the Message
    lixi_message = _Message(message, message_path, file_type)

    # Validate the Message
    isvalid, result, error_logs = lixi_message.validate(schema=schema_text, schema_path=schema_path)

    # A message is returned with error for LIXI API requirement
    if isvalid == False:
        raise LIXIValidationError(result, lixi_message, error_logs)

    return lixi_message


                


# Setting default schema folder as starting folder
# _cwd = _os.getcwd()
# set_schema_folder(_cwd)
# print(_cwd)

###################
## TEST CODE
###################

# la_schema = get_customized_schema(instructions_path='C:/Users/compb/Documents/Git/lixi-pypi/tests/Customisation_Instruction_CAL.xml', schema_path='C:/Users/compb/Documents/Git/lixi-pypi/tests/LIXI-CAL-2_6_19-Annotated.xsd', output_name='Ammar_CAL', output_folder='C:/Users/compb/Documents/Git/lixi-pypi/tests')
# print('hallelujah')
